@@ -12,6 +12,7 @@ import platform
 from datetime import datetime, timezone
 import dateutil.parser
 import ssl
+import re
 
 from .common import ClientError, LoginError, InvalidProject
 from .merginproject import MerginProject
@@ -41,7 +42,7 @@ def decode_token_data(token):
     if not token.startswith(token_prefix):
         raise TokenError(f"Token doesn't start with 'Bearer .': {token}")
     try:
-        data = token[len(token_prefix):].split('.')[0]
+        data = token[len(token_prefix) :].split(".")[0]
         # add proper base64 padding"
         data += "=" * (-len(data) % 4)
         decoded = zlib.decompress(base64.urlsafe_b64decode(data))
@@ -52,24 +53,25 @@ def decode_token_data(token):
 
 class MerginClient:
     """
-    Client for Mergin service.
+    Client for Mergin Maps service.
 
-    :param url: String, Mergin service URL.
-    :param auth_token: String, Mergin authorization token.
-    :param login: String, login for Mergin service.
-    :param password: String, password for Mergin service.
+    :param url: String, Mergin Maps service URL.
+    :param auth_token: String, Mergin Maps authorization token.
+    :param login: String, login for Mergin Maps service.
+    :param password: String, password for Mergin Maps service.
     :param plugin_version: String, info about plugin and QGIS version.
-    :param proxy_config: Dictionary, proxy settings to use when connecting to Mergin service. At least url and port
+    :param proxy_config: Dictionary, proxy settings to use when connecting to Mergin Maps service. At least url and port
         of the server should be provided. Expected keys: "url", "port", "user", "password".
         Currently, only HTTP proxies are supported.
     """
+
     def __init__(self, url=None, auth_token=None, login=None, password=None, plugin_version=None, proxy_config=None):
         self.url = url if url is not None else MerginClient.default_url()
         self._auth_params = None
         self._auth_session = None
         self._user_info = None
         self.client_version = "Python-client/" + __version__
-        if plugin_version is not None:   # this could be e.g. "Plugin/2020.1 QGIS/3.14"
+        if plugin_version is not None:  # this could be e.g. "Plugin/2020.1 QGIS/3.14"
             self.client_version += " " + plugin_version
         self.setup_logging()
         if auth_token:
@@ -89,7 +91,9 @@ class MerginClient:
             if proxy_config["user"] and proxy_config["password"]:
                 handlers.append(
                     urllib.request.ProxyHandler(
-                        {"https": f"{proxy_config['user']}:{proxy_config['password']}@{proxy_url}:{proxy_config['port']}"}
+                        {
+                            "https": f"{proxy_config['user']}:{proxy_config['password']}@{proxy_url}:{proxy_config['port']}"
+                        }
                     )
                 )
                 handlers.append(urllib.request.HTTPBasicAuthHandler())
@@ -104,7 +108,7 @@ class MerginClient:
         if os.path.exists(default_capath):
             self.opener = urllib.request.build_opener(*handlers, urllib.request.HTTPSHandler())
         else:
-            cafile = os.path.join(this_dir, 'cert.pem')
+            cafile = os.path.join(this_dir, "cert.pem")
             if not os.path.exists(cafile):
                 raise Exception("missing " + cafile)
             ctx = ssl.SSLContext()
@@ -119,39 +123,37 @@ class MerginClient:
             raise ClientError("Unable to log in: password provided but no username/email")
 
         if login and password:
-            self._auth_params = {
-                "login": login,
-                "password": password
-            }
+            self._auth_params = {"login": login, "password": password}
             if not self._auth_session:
                 self.login(login, password)
 
     def setup_logging(self):
-        """Setup Mergin client logging."""
-        client_log_file = os.environ.get('MERGIN_CLIENT_LOG', None)
-        self.log = logging.getLogger('mergin.client')
+        """Setup Mergin Maps client logging."""
+        client_log_file = os.environ.get("MERGIN_CLIENT_LOG", None)
+        self.log = logging.getLogger("mergin.client")
         self.log.setLevel(logging.DEBUG)  # log everything (it would otherwise log just warnings+errors)
         if not self.log.handlers:
             if client_log_file:
                 log_handler = logging.FileHandler(client_log_file)
-                log_handler.setFormatter(logging.Formatter('%(asctime)s %(message)s'))
+                log_handler.setFormatter(logging.Formatter("%(asctime)s %(message)s"))
                 self.log.addHandler(log_handler)
             else:
-                # no Mergin log path in the environment - create a null handler that does nothing
+                # no Mergin Maps log path in the environment - create a null handler that does nothing
                 null_logger = logging.NullHandler()
                 self.log.addHandler(null_logger)
 
     @staticmethod
     def default_url():
-        """ Returns URL of the public instance of Mergin """
-        return "https://public.cloudmergin.com"
+        """Returns URL of the public instance of Mergin Maps"""
+        return "https://app.merginmaps.com"
 
     def user_agent_info(self):
-        """ Returns string as it is sent as User-Agent http header to the server """
+        """Returns string as it is sent as User-Agent http header to the server"""
         system_version = "Unknown"
         if platform.system() == "Linux":
             try:
                 from pip._vendor import distro
+
                 system_version = distro.linux_distribution()[0]
             except ModuleNotFoundError:  # pip may not be installed...
                 system_version = "Linux"
@@ -163,6 +165,7 @@ class MerginClient:
 
     def _check_token(f):
         """Wrapper for creating/renewing authorization token."""
+
         def wrapper(self, *args):
             if self._auth_params:
                 if self._auth_session:
@@ -174,8 +177,9 @@ class MerginClient:
                 else:
                     # Create a new authorization token
                     self.log.info(f"No token - login user: {self._auth_params['login']}")
-                    self.login(self._auth_params['login'], self._auth_params['password'])
+                    self.login(self._auth_params["login"], self._auth_params["password"])
             return f(self, *args)
+
         return wrapper
 
     @_check_token
@@ -218,9 +222,9 @@ class MerginClient:
         """
         resp = self.get("/ping")
         data = json.load(resp)
-        if 'endpoints' not in data:
+        if "endpoints" not in data:
             return False
-        endpoints = data['endpoints']
+        endpoints = data["endpoints"]
 
         client_endpoints = {
             "data_sync": {
@@ -229,21 +233,19 @@ class MerginClient:
                     "/project/push/cancel/{transaction_id}",
                     "/project/push/finish/{transaction_id}",
                     "/project/push/{namespace}/{project_name}",
-                   # "/project/push/chunk/{transaction_id}/{chunk_id}" # issue in server
-                ]
+                    # "/project/push/chunk/{transaction_id}/{chunk_id}" # issue in server
+                ],
             },
             "project": {
                 "DELETE": ["/project/{namespace}/{project_name}"],
                 "GET": [
                     "/project",
                     "/project/{namespace}/{project_name}",
-                    "/project/version/{namespace}/{project_name}"
+                    "/project/version/{namespace}/{project_name}",
                 ],
-                "POST": ["/project/{namespace}"]
+                "POST": ["/project/{namespace}"],
             },
-            "user": {
-                "POST": ["/auth/login"]
-            }
+            "user": {"POST": ["/auth/login"]},
         }
 
         for k, v in client_endpoints.items():
@@ -267,10 +269,7 @@ class MerginClient:
         :param password: User's password
         :type password: String
         """
-        params = {
-            "login": login,
-            "password": password
-        }
+        params = {"login": login, "password": password}
         self._auth_session = None
         self.log.info(f"Going to log in user {login}")
         try:
@@ -294,16 +293,14 @@ class MerginClient:
             raise ClientError("failure reason: " + str(e.reason))
         self._auth_session = {
             "token": "Bearer %s" % session["token"],
-            "expire": dateutil.parser.parse(session["expire"])
-        }
-        self._user_info = {
-            "username": data["username"]
+            "expire": dateutil.parser.parse(session["expire"]),
         }
+        self._user_info = {"username": data["username"]}
         self.log.info(f"User {data['username']} successfully logged in.")
         return session
 
     def username(self):
-        """ Returns user name used in this session or None if not authenticated """
+        """Returns user name used in this session or None if not authenticated"""
         if not self._user_info:
             return None  # not authenticated
         return self._user_info["username"]
@@ -327,7 +324,7 @@ class MerginClient:
 
     def create_project(self, project_name, is_public=False, namespace=None):
         """
-        Create new project repository in user namespace on Mergin server.
+        Create new project repository in user namespace on Mergin Maps server.
         Optionally initialized from given local directory.
 
         :param project_name: Project name
@@ -342,10 +339,7 @@ class MerginClient:
         if not self._user_info:
             raise Exception("Authentication required")
 
-        params = {
-            "name": project_name,
-            "public": is_public
-        }
+        params = {"name": project_name, "public": is_public}
         if namespace is None:
             namespace = self.username()
         try:
@@ -358,8 +352,8 @@ class MerginClient:
         """
         Convenience method to create project and push the initial version right after that.
         """
-        if os.path.exists(os.path.join(directory, '.mergin')):
-            raise ClientError('Directory is already assigned to a Mergin project (contains .mergin sub-dir)')
+        if os.path.exists(os.path.join(directory, ".mergin")):
+            raise ClientError("Directory is already assigned to a Mergin Maps project (contains .mergin sub-dir)")
 
         if namespace is None:
             namespace = self.username()
@@ -371,10 +365,11 @@ class MerginClient:
             if mp.inspect_files():
                 self.push_project(directory)
 
-    def paginated_projects_list(self, page=1, per_page=50, tags=None, user=None, flag=None, name=None,
-                                namespace=None, order_params=None):
+    def paginated_projects_list(
+        self, page=1, per_page=50, tags=None, user=None, flag=None, name=None, namespace=None, order_params=None
+    ):
         """
-        Find all available mergin projects.
+        Find all available Mergin Maps projects.
 
         :param tags: Filter projects by tags ('valid_qgis', 'mappin_use', input_use')
         :type tags: List
@@ -423,10 +418,11 @@ class MerginClient:
         projects = json.load(resp)
         return projects
 
-    def projects_list(self, tags=None, user=None, flag=None, q=None):
+    def projects_list(self, tags=None, user=None, flag=None, name=None, namespace=None, order_params=None):
         """
-        Find all available Mergin projects. It will always retrieve max 100 projects.
-        Consider using the paginated_projects_list instead.
+        Find all available Mergin Maps projects.
+
+        Calls paginated_projects_list for all pages. Can take significant time to fetch all pages.
 
         :param tags: Filter projects by tags ('valid_qgis', 'mappin_use', input_use')
         :type tags: List
@@ -437,22 +433,40 @@ class MerginClient:
         :param flag: Predefined filter flag ('created', 'shared')
         :type flag: String
 
-        :param q: Search query string
-        :type q: String
+        :param name: Filter projects with name like name
+        :type name: String
+
+        :param namespace: Filter projects with namespace like namespace
+        :type namespace: String
+
+        :param order_params: optional attributes for sorting the list. It should be a comma separated attribute names
+            with _asc or _desc appended for sorting direction. For example: "namespace_asc,disk_usage_desc".
+            Available attrs: namespace, name, created, updated, disk_usage, creator
+        :type order_params: String
 
         :rtype: List[Dict]
         """
-        params = {}
-        if tags:
-            params["tags"] = ",".join(tags)
-        if user:
-            params["user"] = user
-        if flag:
-            params["flag"] = flag
-        if q:
-            params["q"] = q
-        resp = self.get("/v1/project", params)
-        projects = json.load(resp)
+        projects = []
+        page_i = 1
+        fetched_projects = 0
+        while True:
+            resp = self.paginated_projects_list(
+                page=page_i,
+                per_page=50,
+                tags=tags,
+                user=user,
+                flag=flag,
+                name=name,
+                namespace=namespace,
+                order_params=order_params,
+            )
+            fetched_projects += len(resp["projects"])
+            count = resp["count"]
+            projects += resp["projects"]
+            if fetched_projects < count:
+                page_i += 1
+            else:
+                break
         return projects
 
     def project_info(self, project_path, since=None, version=None):
@@ -467,10 +481,10 @@ class MerginClient:
         :type version: String
         :rtype: Dict
         """
-        params = {'since': since} if since else {}
+        params = {"since": since} if since else {}
         # since and version are mutually exclusive
         if version:
-            params = {'version': version}
+            params = {"version": version}
         resp = self.get("/v1/project/{}".format(project_path), params)
         return json.load(resp)
 
@@ -495,26 +509,18 @@ class MerginClient:
         start_page = math.ceil(num_since / per_page)
         if not num_to:
             # let's get first page and count
-            params = {
-                "page": start_page,
-                "per_page": per_page,
-                "descending": False
-            }
+            params = {"page": start_page, "per_page": per_page, "descending": False}
             resp = self.get("/v1/project/versions/paginated/{}".format(project_path), params)
             resp_json = json.load(resp)
             versions = resp_json["versions"]
-            num_to = resp_json['count']
+            num_to = resp_json["count"]
             latest_version = int_version(versions[-1]["name"])
             if latest_version < num_to:
                 versions += self.project_versions(project_path, f"v{latest_version+1}", f"v{num_to}")
         else:
             end_page = math.ceil(num_to / per_page)
-            for page in range(start_page, end_page+1):
-                params = {
-                    "page": page,
-                    "per_page": per_page,
-                    "descending": False
-                }
+            for page in range(start_page, end_page + 1):
+                params = {"page": page, "per_page": per_page, "descending": False}
                 resp = self.get("/v1/project/versions/paginated/{}".format(project_path), params)
                 versions += json.load(resp)["versions"]
 
@@ -551,7 +557,7 @@ class MerginClient:
         return True, free_space
 
     def user_info(self):
-        resp = self.get('/v1/user/' + self.username())
+        resp = self.get("/v1/user/" + self.username())
         return json.load(resp)
 
     def set_project_access(self, project_path, access):
@@ -566,7 +572,7 @@ class MerginClient:
         params = {"access": access}
         path = "/v1/project/%s" % project_path
         url = urllib.parse.urljoin(self.url, urllib.parse.quote(path))
-        json_headers = {'Content-Type': 'application/json'}
+        json_headers = {"Content-Type": "application/json"}
         try:
             request = urllib.request.Request(url, data=json.dumps(params).encode(), headers=json_headers, method="PUT")
             self._do_request(request)
@@ -585,7 +591,7 @@ class MerginClient:
             raise ClientError("Unsupported permission level")
 
         project_info = self.project_info(project_path)
-        access = project_info.get('access')
+        access = project_info.get("access")
         for name in usernames:
             if permission_level == "owner":
                 access.get("ownersnames").append(name)
@@ -597,12 +603,12 @@ class MerginClient:
 
     def remove_user_permissions_from_project(self, project_path, usernames):
         """
-       Removes specified users from project
-       :param project_path: project full name (<namespace>/<name>)
-       :param usernames: list of usernames to be granted specified permission level
-       """
+        Removes specified users from project
+        :param project_path: project full name (<namespace>/<name>)
+        :param usernames: list of usernames to be granted specified permission level
+        """
         project_info = self.project_info(project_path)
-        access = project_info.get('access')
+        access = project_info.get("access")
         for name in usernames:
             if name in access.get("ownersnames"):
                 access.get("ownersnames").remove(name)
@@ -614,14 +620,14 @@ class MerginClient:
 
     def project_user_permissions(self, project_path):
         """
-       Returns permissions for project
-       :param project_path: project full name (<namespace>/<name>)
-       :return dict("owners": list(usernames),
-                    "writers": list(usernames),
-                    "readers": list(usernames))
-       """
+        Returns permissions for project
+        :param project_path: project full name (<namespace>/<name>)
+        :return dict("owners": list(usernames),
+                     "writers": list(usernames),
+                     "readers": list(usernames))
+        """
         project_info = self.project_info(project_path)
-        access = project_info.get('access')
+        access = project_info.get("access")
         result = {}
         result["owners"] = access.get("ownersnames")
         result["writers"] = access.get("writersnames")
@@ -637,7 +643,7 @@ class MerginClient:
         """
         job = push_project_async(self, directory)
         if job is None:
-            return   # there is nothing to push (or we only deleted some files)
+            return  # there is nothing to push (or we only deleted some files)
         push_project_wait(job)
         push_project_finalize(job)
 
@@ -650,7 +656,7 @@ class MerginClient:
         """
         job = pull_project_async(self, directory)
         if job is None:
-            return   # project is up to date
+            return  # project is up to date
         pull_project_wait(job)
         return pull_project_finalize(job)
 
@@ -667,10 +673,10 @@ class MerginClient:
         """
         path = "/v1/project/clone/%s" % source_project_path
         url = urllib.parse.urljoin(self.url, urllib.parse.quote(path))
-        json_headers = {'Content-Type': 'application/json', 'Accept': 'application/json'}
+        json_headers = {"Content-Type": "application/json", "Accept": "application/json"}
         data = {
-            'namespace': cloned_project_namespace if cloned_project_namespace else self.username(),
-            'project': cloned_project_name
+            "namespace": cloned_project_namespace if cloned_project_namespace else self.username(),
+            "project": cloned_project_name,
         }
 
         request = urllib.request.Request(url, data=json.dumps(data).encode(), headers=json_headers, method="POST")
@@ -710,28 +716,28 @@ class MerginClient:
         return pull_changes, push_changes, push_changes_summary
 
     def project_version_info(self, project_path, version):
-        """ Returns JSON with detailed information about a single project version"""
-        params = {'version_id': version}
+        """Returns JSON with detailed information about a single project version"""
+        params = {"version_id": version}
         resp = self.get("/v1/project/version/{}".format(project_path), params)
         return json.load(resp)
 
     def project_file_history_info(self, project_path, file_path):
-        """ Returns JSON with full history of a single file within a project """
-        params = {'path': file_path}
+        """Returns JSON with full history of a single file within a project"""
+        params = {"path": file_path}
         resp = self.get("/v1/resource/history/{}".format(project_path), params)
         return json.load(resp)
 
     def project_file_changeset_info(self, project_path, file_path, version):
-        """ Returns JSON with changeset details of a particular version of a file within a project """
-        params = {'path': file_path}
+        """Returns JSON with changeset details of a particular version of a file within a project"""
+        params = {"path": file_path}
         resp = self.get("/v1/resource/changesets/{}/{}".format(project_path, version), params)
         return json.load(resp)
 
     def get_projects_by_names(self, projects):
-        """ Returns JSON with projects' info for list of required projects.
+        """Returns JSON with projects' info for list of required projects.
         The schema of the returned information is the same as the response from projects_list().
 
-        This is useful when we have a couple of Mergin projects available locally and we want to
+        This is useful when we have a couple of Mergin Maps projects available locally and we want to
         find out their status at once (e.g. whether there is a new version on the server).
 
         :param projects: list of projects in the form 'namespace/project_name'
@@ -759,7 +765,7 @@ class MerginClient:
         download_file_finalize(job)
 
     def get_file_diff(self, project_dir, file_path, output_diff, version_from, version_to):
-        """ Create concatenated diff for project file diffs between versions version_from and version_to.
+        """Create concatenated diff for project file diffs between versions version_from and version_to.
 
         :param project_dir: project local directory
         :type project_dir: String
@@ -778,7 +784,13 @@ class MerginClient:
         versions_to_fetch = get_versions_with_file_changes(
             self, project_path, file_path, version_from=version_from, version_to=version_to, file_history=file_history
         )
-        diffs = self.download_file_diffs(project_dir, file_path, versions_to_fetch[1:])
+        self.download_file_diffs(project_dir, file_path, versions_to_fetch[1:])
+
+        # collect required versions from the cache
+        diffs = []
+        for v in versions_to_fetch[1:]:
+            diffs.append(mp.fpath_cache(file_history["history"][v]["diff"]["path"], v))
+
         # concatenate diffs, if needed
         output_dir = os.path.dirname(output_diff)
         if len(diffs) >= 1:
@@ -789,7 +801,8 @@ class MerginClient:
                 shutil.copy(diffs[0], output_diff)
 
     def download_file_diffs(self, project_dir, file_path, versions):
-        """ Download file diffs for specified versions.
+        """Download file diffs for specified versions if they are not present
+        in the cache.
 
         :param project_dir: project local directory
         :type project_dir: String
