@@ -1,4 +1,3 @@
-
 import os
 
 
@@ -28,9 +27,10 @@ try:
     from dateutil.tz import tzlocal
 except ImportError:
     # this is to import all dependencies shipped with package (e.g. to use in qgis-plugin)
-    deps_dir = os.path.join(this_dir, 'deps')
+    deps_dir = os.path.join(this_dir, "deps")
     if os.path.exists(deps_dir):
         import sys
+
         for f in os.listdir(os.path.join(deps_dir)):
             sys.path.append(os.path.join(deps_dir, f))
 
