@@ -1,4 +1,3 @@
-
 import json
 import logging
 import math
@@ -19,7 +18,7 @@ from .utils import (
     do_sqlite_checkpoint,
     unique_path_name,
     conflicted_copy_file_name,
-    edit_conflict_file_name
+    edit_conflict_file_name,
 )
 
 
@@ -36,23 +35,24 @@ except (ImportError, ModuleNotFoundError):
 
 
 class MerginProject:
-    """ Base class for Mergin local projects.
+    """Base class for Mergin Maps local projects.
 
     Linked to existing local directory, with project metadata (mergin.json) and backups located in .mergin directory.
     """
+
     def __init__(self, directory):
         self.dir = os.path.abspath(directory)
         if not os.path.exists(self.dir):
-            raise InvalidProject('Project directory does not exist')
+            raise InvalidProject("Project directory does not exist")
 
-        self.meta_dir = os.path.join(self.dir, '.mergin')
+        self.meta_dir = os.path.join(self.dir, ".mergin")
         if not os.path.exists(self.meta_dir):
             os.mkdir(self.meta_dir)
 
         # location for files from unfinished pull
-        self.unfinished_pull_dir = os.path.join(self.meta_dir, 'unfinished_pull')
+        self.unfinished_pull_dir = os.path.join(self.meta_dir, "unfinished_pull")
 
-        self.cache_dir = os.path.join(self.meta_dir, '.cache')
+        self.cache_dir = os.path.join(self.meta_dir, ".cache")
         if not os.path.exists(self.cache_dir):
             os.mkdir(self.cache_dir)
 
@@ -75,18 +75,19 @@ class MerginProject:
                 self.log.warning("GEODIFF: " + text)
             else:
                 self.log.info("GEODIFF: " + text)
+
         self.geodiff.set_logger_callback(_logger_callback)
         self.geodiff.set_maximum_logger_level(pygeodiff.GeoDiff.LevelDebug)
 
     def setup_logging(self, logger_name):
         """Setup logging into project directory's .mergin/client-log.txt file."""
-        self.log = logging.getLogger('mergin.project.' + logger_name)
+        self.log = logging.getLogger("mergin.project." + logger_name)
         self.log.setLevel(logging.DEBUG)  # log everything (it would otherwise log just warnings+errors)
         if not self.log.handlers:
             # we only need to set the handler once
             # (otherwise we would get things logged multiple times as loggers are cached)
-            log_handler = logging.FileHandler(os.path.join(self.meta_dir, "client-log.txt"), encoding = 'utf-8')
-            log_handler.setFormatter(logging.Formatter('%(asctime)s %(message)s'))
+            log_handler = logging.FileHandler(os.path.join(self.meta_dir, "client-log.txt"), encoding="utf-8")
+            log_handler.setFormatter(logging.Formatter("%(asctime)s %(message)s"))
             self.log.addHandler(log_handler)
 
     def remove_logging_handler(self):
@@ -116,15 +117,15 @@ class MerginProject:
         return abs_path
 
     def fpath_meta(self, file):
-        """ Helper function to get absolute path of file in meta dir. """
+        """Helper function to get absolute path of file in meta dir."""
         return self.fpath(file, self.meta_dir)
 
     def fpath_unfinished_pull(self, file):
-        """ Helper function to get absolute path of file in unfinished_pull dir. """
+        """Helper function to get absolute path of file in unfinished_pull dir."""
         return self.fpath(file, self.unfinished_pull_dir)
 
     def fpath_cache(self, file, version=None):
-        """ Helper function to get absolute path of file in cache dir.
+        """Helper function to get absolute path of file in cache dir.
         It can be either in root cache directory (.mergin/.cache/) or in some version's subfolder
         """
         if version:
@@ -133,25 +134,25 @@ class MerginProject:
 
     @property
     def metadata(self):
-        if not os.path.exists(self.fpath_meta('mergin.json')):
-            raise InvalidProject('Project metadata has not been created yet')
-        with open(self.fpath_meta('mergin.json'), 'r') as file:
+        if not os.path.exists(self.fpath_meta("mergin.json")):
+            raise InvalidProject("Project metadata has not been created yet")
+        with open(self.fpath_meta("mergin.json"), "r") as file:
             return json.load(file)
 
     @metadata.setter
     def metadata(self, data):
-        with open(self.fpath_meta('mergin.json'), 'w') as file:
+        with open(self.fpath_meta("mergin.json"), "w") as file:
             json.dump(data, file, indent=2)
 
     def is_versioned_file(self, file):
-        """ Check if file is compatible with geodiff lib and hence suitable for versioning.
+        """Check if file is compatible with geodiff lib and hence suitable for versioning.
 
         :param file: file path
         :type file: str
         :returns: if file is compatible with geodiff lib
         :rtype: bool
         """
-        diff_extensions = ['.gpkg', '.sqlite']
+        diff_extensions = [".gpkg", ".sqlite"]
         f_extension = os.path.splitext(file)[1]
         return f_extension in diff_extensions
 
@@ -165,9 +166,9 @@ class MerginProject:
         :rtype: bool
         """
         f_extension = os.path.splitext(path)[1]
-        if f_extension != '.gpkg':
+        if f_extension != ".gpkg":
             return False
-        if os.path.exists(f'{path}-wal'):
+        if os.path.exists(f"{path}-wal"):
             return True
         return False
 
@@ -180,8 +181,8 @@ class MerginProject:
         :returns: whether file should be ignored
         :rtype: bool
         """
-        ignore_ext = re.compile(r'({})$'.format('|'.join(re.escape(x) for x in ['-shm', '-wal', '~', 'pyc', 'swap'])))
-        ignore_files = ['.DS_Store', '.directory']
+        ignore_ext = re.compile(r"({})$".format("|".join(re.escape(x) for x in ["-shm", "-wal", "~", "pyc", "swap"])))
+        ignore_files = [".DS_Store", ".directory"]
         name, ext = os.path.splitext(file)
         if ext and ignore_ext.search(ext):
             return True
@@ -198,20 +199,22 @@ class MerginProject:
         """
         files_meta = []
         for root, dirs, files in os.walk(self.dir, topdown=True):
-            dirs[:] = [d for d in dirs if d not in ['.mergin']]
+            dirs[:] = [d for d in dirs if d not in [".mergin"]]
             for file in files:
                 if self.ignore_file(file):
                     continue
 
                 abs_path = os.path.abspath(os.path.join(root, file))
                 rel_path = os.path.relpath(abs_path, start=self.dir)
-                proj_path = '/'.join(rel_path.split(os.path.sep))  # we need posix path
-                files_meta.append({
-                    "path": proj_path,
-                    "checksum": generate_checksum(abs_path),
-                    "size": os.path.getsize(abs_path),
-                    "mtime": datetime.fromtimestamp(os.path.getmtime(abs_path), tzlocal())
-                })
+                proj_path = "/".join(rel_path.split(os.path.sep))  # we need posix path
+                files_meta.append(
+                    {
+                        "path": proj_path,
+                        "checksum": generate_checksum(abs_path),
+                        "size": os.path.getsize(abs_path),
+                        "mtime": datetime.fromtimestamp(os.path.getmtime(abs_path), tzlocal()),
+                    }
+                )
         return files_meta
 
     def compare_file_sets(self, origin, current):
@@ -247,12 +250,7 @@ class MerginProject:
             f["origin_checksum"] = origin_map[path]["checksum"]
             updated.append(f)
 
-        return {
-            "renamed": [],
-            "added": added,
-            "removed": removed,
-            "updated": updated
-        }
+        return {"renamed": [], "added": added, "removed": removed, "updated": updated}
 
     def get_pull_changes(self, server_files):
         """
@@ -272,11 +270,11 @@ class MerginProject:
         """
 
         # first let's have a look at the added/updated/removed files
-        changes = self.compare_file_sets(self.metadata['files'], server_files)
+        changes = self.compare_file_sets(self.metadata["files"], server_files)
 
         # then let's inspect our versioned files (geopackages) if there are any relevant changes
         not_updated = []
-        for file in changes['updated']:
+        for file in changes["updated"]:
             if not self.is_versioned_file(file["path"]):
                 continue
 
@@ -286,28 +284,28 @@ class MerginProject:
 
             # get sorted list of the history (they may not be sorted or using lexical sorting - "v10", "v11", "v5", "v6", ...)
             history_list = []
-            for version_str, version_info in file['history'].items():
-                history_list.append( (int_version(version_str), version_info) )
+            for version_str, version_info in file["history"].items():
+                history_list.append((int_version(version_str), version_info))
             history_list = sorted(history_list, key=lambda item: item[0])  # sort tuples based on version numbers
 
             # need to track geodiff file history to see if there were any changes
             for version, version_info in history_list:
-                if version <= int_version(self.metadata['version']):
+                if version <= int_version(self.metadata["version"]):
                     continue  # ignore history of no interest
                 is_updated = True
-                if 'diff' in version_info:
-                    diffs.append(version_info['diff']['path'])
-                    diffs_size += version_info['diff']['size']
+                if "diff" in version_info:
+                    diffs.append(version_info["diff"]["path"])
+                    diffs_size += version_info["diff"]["size"]
                 else:
                     diffs = []
                     break  # we found force update in history, does not make sense to download diffs
 
             if is_updated:
-                file['diffs'] = diffs
+                file["diffs"] = diffs
             else:
                 not_updated.append(file)
 
-        changes['updated'] = [f for f in changes['updated'] if f not in not_updated]
+        changes["updated"] = [f for f in changes["updated"] if f not in not_updated]
         return changes
 
     def get_push_changes(self):
@@ -323,18 +321,18 @@ class MerginProject:
         :returns: changes metadata for files to be pushed to server
         :rtype: dict
         """
-        changes = self.compare_file_sets(self.metadata['files'], self.inspect_files())
+        changes = self.compare_file_sets(self.metadata["files"], self.inspect_files())
         # do checkpoint to push changes from wal file to gpkg
-        for file in changes['added'] + changes['updated']:
+        for file in changes["added"] + changes["updated"]:
             size, checksum = do_sqlite_checkpoint(self.fpath(file["path"]), self.log)
             if size and checksum:
                 file["size"] = size
                 file["checksum"] = checksum
-            file['chunks'] = [str(uuid.uuid4()) for i in range(math.ceil(file["size"] / UPLOAD_CHUNK_SIZE))]
+            file["chunks"] = [str(uuid.uuid4()) for i in range(math.ceil(file["size"] / UPLOAD_CHUNK_SIZE))]
 
         # need to check for for real changes in geodiff files using geodiff tool (comparing checksum is not enough)
         not_updated = []
-        for file in changes['updated']:
+        for file in changes["updated"]:
             path = file["path"]
             if not self.is_versioned_file(path):
                 continue
@@ -343,20 +341,20 @@ class MerginProject:
             current_file = self.fpath(path)
             origin_file = self.fpath_meta(path)
             diff_id = str(uuid.uuid4())
-            diff_name = path + '-diff-' + diff_id
+            diff_name = path + "-diff-" + diff_id
             diff_file = self.fpath_meta(diff_name)
             try:
                 self.geodiff.create_changeset(origin_file, current_file, diff_file)
                 if self.geodiff.has_changes(diff_file):
                     diff_size = os.path.getsize(diff_file)
-                    file['checksum'] = file['origin_checksum']  # need to match basefile on server
-                    file['chunks'] = [str(uuid.uuid4()) for i in range(math.ceil(diff_size / UPLOAD_CHUNK_SIZE))]
-                    file['mtime'] = datetime.fromtimestamp(os.path.getmtime(current_file), tzlocal())
-                    file['diff'] = {
+                    file["checksum"] = file["origin_checksum"]  # need to match basefile on server
+                    file["chunks"] = [str(uuid.uuid4()) for i in range(math.ceil(diff_size / UPLOAD_CHUNK_SIZE))]
+                    file["mtime"] = datetime.fromtimestamp(os.path.getmtime(current_file), tzlocal())
+                    file["diff"] = {
                         "path": diff_name,
                         "checksum": generate_checksum(diff_file),
                         "size": diff_size,
-                        'mtime': datetime.fromtimestamp(os.path.getmtime(diff_file), tzlocal())
+                        "mtime": datetime.fromtimestamp(os.path.getmtime(diff_file), tzlocal()),
                     }
                 else:
                     not_updated.append(file)
@@ -366,7 +364,7 @@ class MerginProject:
                 # we will need to do full upload of the file
                 pass
 
-        changes['updated'] = [f for f in changes['updated'] if f not in not_updated]
+        changes["updated"] = [f for f in changes["updated"] if f not in not_updated]
         return changes
 
     def copy_versioned_file_for_upload(self, f, tmp_dir):
@@ -394,7 +392,7 @@ class MerginProject:
                 result_file = self.fpath("change_list" + str(idx), self.meta_dir)
                 try:
                     self.geodiff.list_changes_summary(changeset, result_file)
-                    with open(result_file, 'r') as f:
+                    with open(result_file, "r") as f:
                         change = f.read()
                         changes[file["path"]] = json.loads(change)
                     os.remove(result_file)
@@ -422,22 +420,22 @@ class MerginProject:
         local_changes = self.get_push_changes()
         modified = {}
         for f in local_changes["added"] + local_changes["updated"]:
-            modified.update({f['path']: f})
+            modified.update({f["path"]: f})
 
         local_files_map = {}
         for f in self.inspect_files():
-            local_files_map.update({f['path']: f})
+            local_files_map.update({f["path"]: f})
 
         for k, v in changes.items():
             for item in v:
-                path = item['path']
+                path = item["path"]
                 src = self.fpath(path, temp_dir)
                 dest = self.fpath(path)
                 basefile = self.fpath_meta(path)
 
                 # special care is needed for geodiff files
                 # 'src' here is server version of file and 'dest' is locally modified
-                if self.is_versioned_file(path) and k == 'updated':
+                if self.is_versioned_file(path) and k == "updated":
                     if path in modified:
                         conflict = self.update_with_rebase(path, src, dest, basefile, temp_dir, user_name)
                         if conflict:
@@ -448,11 +446,11 @@ class MerginProject:
                         self.update_without_rebase(path, src, dest, basefile, temp_dir)
                 else:
                     # backup if needed
-                    if path in modified and item['checksum'] != local_files_map[path]['checksum']:
+                    if path in modified and item["checksum"] != local_files_map[path]["checksum"]:
                         conflict = self.create_conflicted_copy(path, user_name)
                         conflicts.append(conflict)
 
-                    if k == 'removed':
+                    if k == "removed":
                         if os.path.exists(dest):
                             os.remove(dest)
                         else:
@@ -492,15 +490,15 @@ class MerginProject:
         """
         self.log.info("updating file with rebase: " + path)
 
-        server_diff = self.fpath(f'{path}-server_diff', temp_dir)  # diff between server file and local basefile
-        local_diff = self.fpath(f'{path}-local_diff', temp_dir)
+        server_diff = self.fpath(f"{path}-server_diff", temp_dir)  # diff between server file and local basefile
+        local_diff = self.fpath(f"{path}-local_diff", temp_dir)
 
         # temporary backup of file pulled from server for recovery
-        f_server_backup = self.fpath(f'{path}-server_backup', temp_dir)
+        f_server_backup = self.fpath(f"{path}-server_backup", temp_dir)
         self.geodiff.make_copy_sqlite(src, f_server_backup)
 
         # create temp backup (ideally with geodiff) of locally modified file if needed later
-        f_conflict_file = self.fpath(f'{path}-local_backup', temp_dir)
+        f_conflict_file = self.fpath(f"{path}-local_backup", temp_dir)
 
         try:
             self.geodiff.create_changeset(basefile, dest, local_diff)
@@ -514,7 +512,8 @@ class MerginProject:
         # they will be stored in a JSON file - if there are no conflicts, the file
         # won't even be created
         rebase_conflicts = unique_path_name(
-                edit_conflict_file_name(self.fpath(path), user_name, int_version(self.metadata['version'])))
+            edit_conflict_file_name(self.fpath(path), user_name, int_version(self.metadata["version"]))
+        )
 
         # try to do rebase magic
         try:
@@ -538,7 +537,7 @@ class MerginProject:
                 f_server_unfinished = self.fpath_unfinished_pull(path)
                 self.geodiff.make_copy_sqlite(f_server_backup, f_server_unfinished)
 
-        return ''
+        return ""
 
     def update_without_rebase(self, path, src, dest, basefile, temp_dir):
         """
@@ -562,7 +561,7 @@ class MerginProject:
         """
         self.log.info("updating file without rebase: " + path)
         try:
-            server_diff = self.fpath(f'{path}-server_diff', temp_dir)  # diff between server file and local basefile
+            server_diff = self.fpath(f"{path}-server_diff", temp_dir)  # diff between server file and local basefile
             # TODO: it could happen that basefile does not exist.
             # It was either never created (e.g. when pushing without geodiff)
             # or it was deleted by mistake(?) by the user. We should detect that
@@ -587,16 +586,16 @@ class MerginProject:
         """
         for k, v in changes.items():
             for item in v:
-                path = item['path']
+                path = item["path"]
                 if not self.is_versioned_file(path):
                     continue
 
                 basefile = self.fpath_meta(path)
-                if k == 'removed':
+                if k == "removed":
                     os.remove(basefile)
-                elif k == 'added':
+                elif k == "added":
                     self.geodiff.make_copy_sqlite(self.fpath(path), basefile)
-                elif k == 'updated':
+                elif k == "updated":
                     # in case for geopackage cannot be created diff (e.g. forced update with committed changes from wal file)
                     if "diff" not in item:
                         self.log.info("updating basefile (copy) for: " + path)
@@ -604,7 +603,7 @@ class MerginProject:
                     else:
                         self.log.info("updating basefile (diff) for: " + path)
                         # better to apply diff to previous basefile to avoid issues with geodiff tmp files
-                        changeset = self.fpath_meta(item['diff']['path'])
+                        changeset = self.fpath_meta(item["diff"]["path"])
                         patch_error = self.apply_diffs(basefile, [changeset])
                         if patch_error:
                             # in case of local sync issues it is safier to remove basefile, next time it will be downloaded from server
@@ -626,7 +625,9 @@ class MerginProject:
         if not os.path.exists(src):
             return
 
-        backup_path = unique_path_name(conflicted_copy_file_name(self.fpath(file), user_name, int_version(self.metadata['version'])))
+        backup_path = unique_path_name(
+            conflicted_copy_file_name(self.fpath(file), user_name, int_version(self.metadata["version"]))
+        )
 
         if self.is_versioned_file(file):
             self.geodiff.make_copy_sqlite(src, backup_path)
@@ -660,7 +661,7 @@ class MerginProject:
         return error
 
     def has_unfinished_pull(self):
-        """ Check if there is an unfinished pull for this project.
+        """Check if there is an unfinished pull for this project.
 
         Unfinished pull means that a previous pull_project() call has
         failed in the final stage due to some files being in read-only
@@ -730,3 +731,17 @@ class MerginProject:
         shutil.rmtree(self.unfinished_pull_dir)
         self.log.info("unfinished pull resolved successfuly!")
         return conflicts
+
+    def set_tables_to_skip(self, tables):
+        """
+        Set list of tables to exclude from geodiff operations. Once defined, these
+        tables will be excluded from the following operations: create changeset,
+        apply changeset, rebase, get database schema, dump database contents, copy
+        database between different drivers.
+
+        If empty list is passed, list will be reset.
+
+        :param tables: list of table names to ignore
+        :type tables: list[str]
+        """
+        self.geodiff.set_tables_to_skip(tables)
